/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Vector;

import org.ow2.opensuit.xmlmap.annotations.XmlAncestor;
import org.ow2.opensuit.xmlmap.annotations.XmlAttribute;
import org.ow2.opensuit.xmlmap.annotations.XmlChild;
import org.ow2.opensuit.xmlmap.annotations.XmlChildren;
import org.ow2.opensuit.xmlmap.annotations.XmlContent;
import org.ow2.opensuit.xmlmap.annotations.XmlContentAssist;
import org.ow2.opensuit.xmlmap.annotations.XmlDoc;
import org.ow2.opensuit.xmlmap.annotations.XmlElement;
import org.ow2.opensuit.xmlmap.annotations.XmlEnumeration;
import org.ow2.opensuit.xmlmap.annotations.XmlSimpleType;
import org.ow2.opensuit.xmlmap.annotations.XmlSubstitutionGroup;
import org.ow2.opensuit.xmlmap.interfaces.IContentAssist;
import org.ow2.opensuit.xmlmap.mapping.FieldMapping;
import org.ow2.opensuit.xmlmap.mapping.IMappingMethod;
import org.ow2.opensuit.xmlmap.mapping.MappingError;

/**
 * This class implements <code>IMappingMethod</code> interface.
 * This implementation allow to generated the Java & XML data binding,
 * thank to Java annotations. It's the default mapping method of
 * OpenSUIT Framework.
 * At the moment just three {@link IMappingMethod} exsist:
 * <ul>
 * <li>method-annotations</li>
 * <li>method-declare</li>
 * <li>method-naming</li>
 * </ul>
 * @author Adrien Ruffie / Pierre Smeyers
 * @author http://opensuit.ow2.org/
 * @version 1.0
 */
public class BindWithAnnotations extends IMappingMethod 
{
	// =====================================================================================
	// === Private Stuff
	// =====================================================================================
	private static int jvmFieldEnumerationDirection = 0;// unknown
	private static class TestClass
	{
		private Object field1;
		private Object field2;
	}
	
	/**
	 * Returns whether the JVM enumerates the declared fields in the same order
	 * than they were declared, or in the opposite order.
	 * @return boolean
	 */
	private static boolean fieldEnumFollowsDeclaration()
	{
		if(jvmFieldEnumerationDirection == 0)
		{
			// --- have to init
			Field[] fields = TestClass.class.getDeclaredFields();
			jvmFieldEnumerationDirection = fields[0].getName().equals("field1") ? 1 : -1;
		}
		return jvmFieldEnumerationDirection == 1;
	}
	// =====================================================================================
	// === Type Methods
	// =====================================================================================
	@Override
    /**
     * Allow to check if the provided class parameter is an simple type.
     * Returns true if an <code>XmlSimpleType</code> annotation for the specified 
     * class is present on this element, else false.
     * @param c the Class object which must be checked
     * @return boolean
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#isSimpleType(Class)
     */
	public boolean isSimpleType(Class c)
	{
		return c.isAnnotationPresent(XmlSimpleType.class);
	}
	
	/**
	 * Allow to check if the provided class parameter is an enumeration.
     * Returns true if an <code>XmlEnumeration</code> annotation for the specified 
     * class is present on this element, else false.
     * @param c the Class object which must be checked
     * @return boolean
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#isEnumeration(Class)
	 */
	public boolean isEnumeration(Class c)
	{
		return c.isAnnotationPresent(XmlEnumeration.class);
	}
	
	 /**
     * Allow to check if the provided class parameter importable by
     * checking substitution group annotation. Returns true if an
     * <code>XmlSubstitutionGroup</code> annotation for the specified class is
     * present on this element, else false.
     * @param c the Class object which must be checked
     * @return boolean
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#isImportable(Class)
     */
	public boolean isImportable(Class c)
	{
		XmlSubstitutionGroup annot = (XmlSubstitutionGroup)c.getAnnotation(XmlSubstitutionGroup.class);
		
		if(annot == null){
			return false;
		}
		
		return annot.importable();
	}
	
	 /**
     * Allow to check if the provided class parameter is marked as a
     * substitution group. Returns true if an <code>XmlSubstitutionGroup</code>
     * annotation for the specified class is present on this element,
     * else false.
     * @param c the Class object which must be checked
     * @return boolean
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#isSubstitutionGroup(Class)
     */
	public boolean isSubstitutionGroup(Class c)
	{
		return c.isAnnotationPresent(XmlSubstitutionGroup.class);
	}
	
	 /**
     * Allow to check if the provided class parameter is marked as a
     * <code>XmlElement</code> element. Returns true if an <code>XmlElement</code>
     * annotation for the specified class is present on this element,
     * else false.
     * @param c the Class object which must be checked
     * @return boolean
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#isXmlElement(Class)
     */
	public boolean isXmlElement(Class c)
	{
		return c.isAnnotationPresent(XmlElement.class);
	}
	// =====================================================================================
	// === Simple Type Methods
	// =====================================================================================
	@Override
	/**
     * Allow to retrieve the base type of the specified class function
     * parameter. This provided parameter must be annotated with
     * <code>XmlSimpleType</code> annotation, in order to retrieve the base
     * type
     * @param simpleType the Class object which must be checked
     * @return Class the base class of the specified simple type
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getBaseSimpleType(Class)
     */
	public Class getBaseSimpleType(Class simpleType)
	{
		XmlSimpleType annot = (XmlSimpleType)simpleType.getAnnotation(XmlSimpleType.class);
		
		if(annot == null){
			return null;
		}
		
		return annot.base();
	}
	// =====================================================================================
	// === Element Methods
	// =====================================================================================
    /**
     * Allow populate the mapping list provided in function parameter.
     * This populating is recursivly performed for super class.
     * @param mappedFields fields list which must be populate.
     * @param mappingErrors list where occured errors are stored.
     * @param iClass Class to extract Xml Mappings from.
     * @param iExcludeFrom Class object allowing to check if the
     * super class isn't an exclude type.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#populateMappings(List, List , Class, Class);
     */
	protected void populateMappings(List mappedFields, List mappingErrors, Class iClass, Class iExcludeFrom)
	{
		if(iClass.getSuperclass() != null && iClass.getSuperclass() != iExcludeFrom){
			populateMappings(mappedFields, mappingErrors, iClass.getSuperclass(), iExcludeFrom);
		}
		
		Field[] fields = iClass.getDeclaredFields();
		boolean fieldsInDeclOrder = fieldEnumFollowsDeclaration();
		for(int i=0; i<fields.length; i++)
		{
            try
            {
            	int idx = fieldsInDeclOrder ? i : fields.length - i - 1;
            	FieldMapping mapping = getXmlMapping(fields[idx]);
				
            	if(mapping != null){
					mappedFields.add(mapping);
				}
            }
            catch(MappingError e)
            {
            	mappingErrors.add(e);
            }
		}
	}
	
	 /**
     * Allow to retrieve the field mapping for the specified 
     * function parameter.
     * @param Field the specified field where the mapping must
     * be extract.
     * @return FieldMapping the extracted mapping
     * @throws MappingError can be throws if the field type
     * for an xml attribute is an array
     */
	protected static FieldMapping getXmlMapping(Field field) throws MappingError
	{
		Annotation annot = null;
		if((annot=field.getAnnotation(XmlAttribute.class)) != null)
		{
			// --- Field mapped to an XML attribute
			if(field.getType().isArray() && field.getType() != byte[].class){
				throw new MappingError("XmlMapping error in field '"+field.getDeclaringClass().getName()+"/"+field.getName()+"': mapping to an xmlattr involves the field type is not an array.");
			}
			
			// TODO: check type is supported as an attribute (simple types and enumerations)
			
			XmlAttribute xannot = (XmlAttribute)annot;
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.ATTRIBUTE);
			ret.setName("*".equals(xannot.name()) ? field.getName() : xannot.name());
			ret.setMinOccurs(xannot.required() ? 1 : 0);
			ret.setMaxOccurs(1);
			return ret;
		}
		else if((annot=field.getAnnotation(XmlChild.class)) != null)
		{
			// --- Field mapped to an XML child element
			// TODO: check type is supported as an XML child (simple types, enumerations, substitution groups and xml elements)
			XmlChild xannot = (XmlChild)annot;
			FieldMapping ret = new FieldMapping(field);
			ret.setType(xannot.direct() ? FieldMapping.DIRECT_CHILD_ELT : FieldMapping.CHILD_ELT);
			ret.setName("*".equals(xannot.name()) ? field.getName() : xannot.name());
			ret.setMinOccurs(xannot.required() ? 1 : 0);
			ret.setMaxOccurs(1);
			return ret;
		}
		else if((annot=field.getAnnotation(XmlChildren.class)) != null)
		{
			// --- Field mapped to an XML child element
			// TODO: check type is supported as a XML children (vector type of substitution groups and xml elements)
			XmlChildren xannot = (XmlChildren)annot;
			FieldMapping ret = new FieldMapping(field);
			ret.setType(xannot.direct() ? FieldMapping.DIRECT_CHILD_ELT : FieldMapping.CHILD_ELT);
			ret.setName("*".equals(xannot.name()) ? field.getName() : xannot.name());
			ret.setMinOccurs(xannot.minOccurs());
			ret.setMaxOccurs(xannot.maxOccurs());
			return ret;
		}
		else if(field.isAnnotationPresent(XmlContent.class))
		{
//			if(field.getType() != String.class)
//				throw new MappingError("XmlMapping error in field '"+field.getDeclaringClass().getName()+"/"+field.getName()+"': xmlcontent field has to be of type String.");
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.CONTENT);
			return ret;
		}
		else if(field.isAnnotationPresent(XmlAncestor.class))
		{
			// --- ancestor field must be at least IXmlObject
//			if(!IXmlObject.class.isAssignableFrom(field.getType()))
//				throw new XmlMappingException("XmlMapping error in field '"+field.getDeclaringClass().getName()+"/"+field.getName()+"': xmlancestor field should necessarily be a subclass of IXmlObject.");
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.ANCESTOR);
			return ret;
		}
//		else if(field.isAnnotationPresent(XmlParent.class))
//		{
//			// --- parent field must be at least IXmlObject
//			FieldMapping ret = new FieldMapping(field);
//			ret.setType(FieldMapping.PARENT);
//			return ret;
//		}
//		else if(field.isAnnotationPresent(XmlRoot.class))
//		{
//			// --- root field must be at least IXmlObject
//			FieldMapping ret = new FieldMapping(field);
//			ret.setType(FieldMapping.ROOT);
//			return ret;
//		}
		else
		{
			return null;
		}
	}
	// =====================================================================================
	// === Enumeration Methods
	// =====================================================================================
	private Map<Class<?>, Field[]> class2Fields = new HashMap<Class<?>, Field[]>();
	
	/**
	 * Returns the number of elements in the enumeration.
	 * @param Class represent the enumeration class.
	 * @return int The number of elements.
	 * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getNbOfEnumItems(Class)
     */
	public int getNbOfEnumItems(Class iEnumClass)
	{
		if(iEnumClass.isEnum()){
			return iEnumClass.getEnumConstants().length;
		}
		
		return getFakeEnumItems(iEnumClass).length;
	}
	/**
	 * Returns an enumeration element from its name.
     * @param Class represent the enumeration class.
	 * @param iName The enumeration element name.
	 * @return Object The enumeration element value.
	 * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getEnumItem(Class, String)
	 */
	public Object getEnumItem(Class iEnumClass, String iName)
	{
		if(iEnumClass.isEnum()){
			return Enum.valueOf(iEnumClass, iName);
		}
		
		Field[] elements = getFakeEnumItems(iEnumClass);
		for(int i=0; i<elements.length; i++)
		{
			if(elements[i].getName().equals(iName))
			{
	            try
                {
	                return elements[i].get(null);
                }
                catch(Exception e)
                {
                	System.out.println("Enum Error:");
	                e.printStackTrace();
	                return null;
                }
			}
		}
		return null;
	}
	
	/**
	 * Returns an enumeration element from its index.
	 * @param iIndex The enumeration element index.
     * @param iEnumClass represent the enumeration class.
	 * @return Object The enumeration element value.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getEnumItem(Class, int)
	 */
	public Object getEnumItem(Class iEnumClass, int iIndex)
	{
		if(iEnumClass.isEnum()){
			return iEnumClass.getEnumConstants()[iIndex];
		}
		
		Field[] elements = getFakeEnumItems(iEnumClass);
		
		if(iIndex >= elements.length){
			return null;
		}
		
		try
        {
	        return elements[iIndex].get(null);
        }
        catch(Exception e)
        {
        	System.out.println("Enum Error:");
            e.printStackTrace();
            return null;
        }
	}
	
	/**
	 * Returns an enumeration element name from its index.
	 * @param iIndex The enumeration element index.
	 * @return String The enumeration element name.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getEnumItemName(Class, int)
	 */
	public String getEnumItemName(Class iEnumClass, int iIndex)
	{
		if(iEnumClass.isEnum()){
			return ((Enum)iEnumClass.getEnumConstants()[iIndex]).name();
		}
		
		Field[] elements = getFakeEnumItems(iEnumClass);
		
		if(iIndex >= elements.length){
			return null;
		}
		
		return elements[iIndex].getName();
	}
	
	/**
     * Returns an enumeration name.
     * @param iEnumValue The enumeration element.
     * @return String The enumeration element name.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getEnumItemName(Object)
     */
	public String getEnumItemName(Object iEnumValue)
	{
		if(iEnumValue instanceof Enum<?>){
			return ((Enum<?>)iEnumValue).name();
		}

		Field[] elements = getFakeEnumItems(iEnumValue.getClass());
		for(int i=0; i<elements.length; i++)
		{
			try
            {
	            if(iEnumValue == elements[i].get(null)){
	            	return elements[i].getName();
	            }
            }
            catch(Exception e)
            {
            }
		}
		return null;
	}

	/*
	 * Computes Enumeration elements (by introspection).
	 */
	private Field[] getFakeEnumItems(Class<?> iEnumClass)
	{
		Field[] elements = (Field[])class2Fields.get(iEnumClass);
		if(elements == null)
		{
			Vector elementFields = new Vector();
			Field[] fields = iEnumClass.getDeclaredFields();
			for(int i=0; i<fields.length; i++)
			{
				// --- field must be static and of same type as this
				if((fields[i].getModifiers() & Modifier.STATIC) != 0 && fields[i].getType().equals(iEnumClass)){
					elementFields.addElement(fields[i]);
				}
			}
			elements = new Field[elementFields.size()];
			elementFields.copyInto(elements);
			class2Fields.put(iEnumClass, elements);
		}
		return elements;
	}
	
	// =====================================================================================
	// === Documentation Methods
	// =====================================================================================
    /**
     * Returns the name of the specified element instance and 
     * taking in account the local.
     * @param element instance where the name must be retrieve.
     * @param locale the specified local.
     * @return String The instance name.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getInstanceName(Locale, Object)
     */
	public String getInstanceName(Locale locale, Object element)
	{
		return getElementTitle(locale, element.getClass());
	}
	
	/**
     * Returns the name of the specified class and 
     * taking in account the local.
     * @param iComponentClass the class where
     * the name must be retrieve.
     * @param iLocale the specified local.
     * @return String The instance name.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getElementTitle(Locale, Class)
     */
	public String getElementTitle(Locale iLocale, Class iComponentClass)
	{
		return iComponentClass.getName();
	}
	
	/**
     * Returns detailed documentation of the specified class and 
     * taking in account the local.
     * @param iComponentClass the class where the name
     * must be retrieve.
     * @param iLocale the specified local.
     * @return String documentation value.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getElementDetails(Locale, Class)
     */
	public String getElementDetails(Locale iLocale, Class iComponentClass)
	{
		XmlDoc annot = (XmlDoc)iComponentClass.getAnnotation(XmlDoc.class);
		
		if(annot == null){
			return null;
		}
		
		return annot.value();
	}
	
	/**
     * Returns the mapping title for the specified fieldmapping, 
     * taking in account the local.
     * @param locale the specified local.
     * @param mapping the field mapping where the name must
     * be extract.
     * @return String the mapping name.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getMappingTitle(Locale, FieldMapping)
     */
	public String getMappingTitle(Locale locale, FieldMapping mapping)
	{
		return mapping.getName();
	}
	
	/**
     * Returns detailed documentation of the specified mapping 
     * taking in account the local.
     * @param locale the specified local.
     * @param mapping the field mapping where the documentation
     * value must be extract.
     * @return String the mapping documentation.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getMappingDetails(Locale, FieldMapping)
     */
	public String getMappingDetails(Locale locale, FieldMapping mapping)
	{
		XmlDoc annot = (XmlDoc)mapping.getField().getAnnotation(XmlDoc.class);
		
		if(annot == null){
			return null;
		}
		
		return annot.value();
	}
	
	/**
     * Returns detailed documentation of the specified enumeration 
     * item, taking in account the local.
     * @param locale the specified local.
     * @param enumClass the enumeration class, where the documentation
     * value of the specified enumeration item must be extract.
     * @param name the name of the enumeration item
     * @return String the enumeration item documentation.
     * @see org.ow2.opensuit.xmlmap.mapping.IMappingMethod#getEnumItemDetails(Locale, Class, String)
     */
	public String getEnumItemDetails(Locale locale, Class enumClass, String name)
	{
		try
		{
			Field f = enumClass.getDeclaredField(name);
			XmlDoc annot = (XmlDoc)f.getAnnotation(XmlDoc.class);
			
			if(annot == null){
				return null;
			}
			
			return annot.value();
		}
		catch (SecurityException e)
		{
			return name;
		}
		catch (NoSuchFieldException e)
		{
			return name;
		}
	}
	
	// =====================================================================================
	// === Content Assist Methods
	// =====================================================================================
	public IContentAssist getContentAssist(Object obj, FieldMapping mapping)
	{
		XmlContentAssist annot = (XmlContentAssist)mapping.getField().getAnnotation(XmlContentAssist.class);
		System.out.println("content assist on ");
		if(annot == null){
			// --- there may be content assist on the field class
			annot = (XmlContentAssist)mapping.getField().getType().getAnnotation(XmlContentAssist.class);
			if(annot == null){
				return null;
			}
		}
		
		try {
			Class<? extends IContentAssistHolder> holderClass = annot.value();
			Constructor<? extends IContentAssistHolder> cs = holderClass.getDeclaredConstructor();
//			IContentAssistHolder holder = annot.value().newInstance();
			cs.setAccessible(true);
			IContentAssistHolder holder = cs.newInstance();

			holder.setObject(obj);
			return holder;
		} catch (Exception e) {
System.err.println("error while instanatiating content assist holder: ");
e.printStackTrace();
			return null;
		}
	}
}
